#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from os import (
    path as os_path, remove as os_remove, removedirs as os_removedirs,
)
from unidecode import unidecode

from django.conf import settings
from django.contrib.auth.models import Permission
from django.db import models
from django.utils.crypto import get_random_string
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from proplan.conf import (
    ATTACH_UPLOAD_PATH, ATTACH_THUMB_EXTENSIONS, PRIORITIES,
)


def upload_attach(instance, filename):
    filename = unidecode(filename).lower()
    filename = filename.replace(' ', '_').replace("'", '').replace('"', '')
    dic = {
        'filename': filename,
        'date': now().date().isoformat(),
        'code': get_random_string(),
        'user_id': instance.user_id or 0,
    }
    return ATTACH_UPLOAD_PATH % dic


class AttachmentManager(models.Manager):
    def get_free(self):
        qs = self.get_queryset()
        qs = qs.filter(thread__isnull=True, comment__isnull=True)
        return qs


class Attachment(models.Model):
    """Model of uploaded files."""

    created = models.DateTimeField(
        _('created'), auto_now_add=True, db_index=True)
    updated = models.DateTimeField(
        _('updated'), auto_now=True, db_index=True)
    # Who uploaded this file. The Automatic Bugs System sets as None.
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        blank=True, null=True,
        verbose_name=_('user'),
        related_name='proplan_attaches')
    file = models.FileField(_('file'), upload_to=upload_attach)

    objects = AttachmentManager()

    class Meta:
        verbose_name = _('attach')
        verbose_name_plural = _('attaches')
        ordering = ('-updated', '-created')

    def __str__(self):
        return self.file.path

    def delete(self, **kwargs):
        """Clears the directory of files before deleting the object."""
        if self.use_thumbnail:
            try:
                os_remove(self.thumb_path)
            except OSError:
                pass
        file = self.file
        path = file.path
        file.delete(False)
        try:
            os_removedirs(os_path.dirname(path))
        except OSError:
            pass
        super(Attachment, self).delete(**kwargs)

    @staticmethod
    def get_thumb_filename(filename):
        """Returns the name of the thumbnail created from the file name."""
        return '%s.thumb%s' % os_path.splitext(filename)

    @property
    def is_abs(self):
        """True if it uploaded from automatic bug system."""
        return not self.user_id

    @property
    def extension(self):
        """Returns extension for file in lower case."""
        root, ext = os_path.splitext(self.file.name)
        return ext.lower()

    @property
    def is_image(self):
        """Returns True if the file is an image."""
        L = (
            '.jpg', '.jpeg', '.png', '.bmp', '.gif', '.svg',
        )
        return self.extension in L

    @property
    def use_thumbnail(self):
        """Returns True if the file is an image with thumbnail."""
        return self.extension in ATTACH_THUMB_EXTENSIONS

    @property
    def name(self):
        """Returns the cleaned name of file without path."""
        return os_path.basename(self.file.name)

    @property
    def url(self):
        return self.file.url

    @property
    def thumb_url(self):
        if self.use_thumbnail:
            return self.get_thumb_filename(self.file.url)
        return ''

    @property
    def path(self):
        return self.file.path

    @property
    def thumb_path(self):
        if self.use_thumbnail:
            return self.get_thumb_filename(self.file.path)
        return ''


class TrackerManager(models.Manager):
    def get_abs(self):
        qs = self.get_queryset()
        return qs.filter(is_abs=True).first()


class Tracker(models.Model):
    """
    Model of one tracker. By default exists 'idea', 'feature' and 'bug'.
    """
    name = models.CharField(_('name'), max_length=10, db_index=True)
    description = models.TextField(_('description'), blank=True)
    is_abs = models.BooleanField(
        _('for automatic bug system'), default=False,
        help_text=_('Only one record can be used for Automatic Bug System.'))

    objects = TrackerManager()

    class Meta:
        ordering = ('id',)
        verbose_name = _('tracker')
        verbose_name_plural = _('trackers')

    def __str__(self):
        return self.name

    def save(self, **kwargs):
        super(Tracker, self).save(**kwargs)
        if self.is_abs:
            qs = Tracker.objects.filter(is_abs=True).exclude(id=self.id)
            qs.update(is_abs=False)


class StageManager(models.Manager):
    def get_abs(self):
        qs = self.get_queryset()
        return qs.filter(is_abs=True).first()


class Stage(models.Model):
    """
    Model of one stage. By default exists: 'new', 'discuss', 'develop',
    'testing', 'bugfix', 'finish' and 'reject'.
    """
    name = models.CharField(_('name'), max_length=10, db_index=True)
    description = models.TextField(_('description'), blank=True)
    # As example finished stages: "closed", "rejected", etc...
    is_finished = models.BooleanField(_('is finished'), default=False)
    is_abs = models.BooleanField(
        _('for automatic bug system'), default=False,
        help_text=_(
            'Only one record can be used for new stage for Automatic Bug '
            'System.'
        )
    )

    objects = StageManager()

    class Meta:
        ordering = ('id',)
        verbose_name = _('stage')
        verbose_name_plural = _('stages')

    def __str__(self):
        return self.name

    def save(self, **kwargs):
        # Can't be both as finished and as for ABS.
        if self.is_finished and self.is_abs:
            self.is_finished = False
        super(Stage, self).save(**kwargs)
        if self.is_abs:
            qs = Stage.objects.filter(is_abs=True).exclude(id=self.id)
            qs.update(is_abs=False)


class Role(models.Model):
    """
    Model of one user role.
    """
    name = models.CharField(_('name'), max_length=50, db_index=True)
    description = models.TextField(_('description'), blank=True)
    users = models.ManyToManyField(
        settings.AUTH_USER_MODEL, blank=True, verbose_name=_('users'))
    permissions = models.ManyToManyField(
        Permission,
        verbose_name=_('permissions'),
        blank=True,
        related_name='proplan_roles',
        limit_choices_to={'content_type__app_label': 'proplan'},
    )

    class Meta:
        ordering = ('name',)
        verbose_name = _('user role')
        verbose_name_plural = _('user roles')

    def __str__(self):
        return self.name


class Thread(models.Model):
    """
    Model of one thread.
    """
    created = models.DateTimeField(_('created'), auto_now_add=True)
    updated = models.DateTimeField(_('updated'), auto_now=True)
    # Author of an idea or a bug found. The Automatic Bugs System sets as None.
    author = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        blank=True, null=True,
        verbose_name=_('author'),
        related_name='proplan_author_threads',
    )
    # The parent thread.
    parent = models.ForeignKey(
        'Thread',
        on_delete=models.SET_NULL,
        blank=True, null=True,
        related_name='children',
        verbose_name=_('parent thread'),
    )
    # Tracker is an idea, a bug or something else.
    tracker = models.ForeignKey(Tracker, verbose_name=_('tracker'),
                                on_delete=models.CASCADE)
    # The stage of current.
    stage = models.ForeignKey(Stage, verbose_name=_('stage'),
                              on_delete=models.CASCADE)
    # The higher the priority, the more important it is for the project.
    priority = models.SmallIntegerField(
        _('priority'), choices=PRIORITIES, default=1, db_index=True)
    # True if it received from automatic bug system.
    is_abs = models.BooleanField(
        _('from automatic bug system'), default=False, db_index=True)
    # Title of the thread.
    title = models.CharField(_('title'), max_length=255, db_index=True)
    # Detailed description of the thread.
    message = models.TextField(_('message body'), blank=True, db_index=True)
    # The files attached by the author of the thread.
    attachments = models.ManyToManyField(
        Attachment, blank=True, verbose_name=_('attachments'))

    class Meta:
        verbose_name = _('thread')
        verbose_name_plural = _('threads')
        ordering = ('-updated', '-created')

    def __str__(self):
        return self.title or 'new'

    @property
    def time_total(self):
        """Returns the seconds spended for total work."""
        qs = Executor.objects.filter(thread=self)
        seconds = sum([e.time_total for e in qs])
        return seconds

    @property
    def time_work(self):
        """Returns the seconds spended for work on sprints."""
        qs = Executor.objects.filter(thread=self)
        seconds = sum([e.time_work for e in qs])
        return seconds


class Comment(models.Model):
    """
    Model of one comment for thread.
    """
    created = models.DateTimeField(_('created'), auto_now_add=True)
    updated = models.DateTimeField(_('updated'), auto_now=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        verbose_name=_('user'),
        related_name='proplan_comments',
    )
    thread = models.ForeignKey(
        Thread, on_delete=models.CASCADE, verbose_name=_('thread'))
    message = models.TextField(_('message body'), db_index=True)
    # JSON data about upgrading the thread.
    changes = models.TextField(_('thread changes'), blank=True, editable=False)
    attachments = models.ManyToManyField(
        Attachment, blank=True, verbose_name=_('attachments'))

    class Meta:
        verbose_name = _('comment')
        verbose_name_plural = _('comments')
        ordering = ('-updated', '-created')

    def __str__(self):
        return self.title


class Executor(models.Model):
    """
    Model of one executor for thread.
    """
    created = models.DateTimeField(_('created'), auto_now_add=True)
    updated = models.DateTimeField(_('updated'), auto_now=True)
    thread = models.ForeignKey(
        Thread,
        on_delete=models.PROTECT,
        related_name='executors',
        verbose_name=_('thread'),
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        verbose_name=_('user'),
        related_name='proplan_executors',
    )
    role = models.ForeignKey(
        Role,
        on_delete=models.PROTECT,
        blank=True, null=True,
        verbose_name=_('user role'),
        related_name='executors',
    )
    # Start working on the thread.
    start = models.DateTimeField(
        _('start working'), null=True, blank=True, editable=False,
        db_index=True)
    # End working on the thread.
    stop = models.DateTimeField(
        _('stop working'), null=True, blank=True, editable=False,
        db_index=True)
    # The start time for the current sprint. This should be absent when there
    # is no work.
    sprint = models.DateTimeField(
        _('start of the sprint'), null=True, blank=True, editable=False)
    # The time of sprints in the seconds.
    time_sprints = models.IntegerField(
        _('time of the sprints'), default=0, editable=False, db_index=True)
    # The commentaries of the user between start and end work.
    comments = models.ManyToManyField(
        Comment, blank=True, verbose_name=_('comments'))

    class Meta:
        verbose_name = _('executor')
        verbose_name_plural = _('executors')
        ordering = ('-updated', '-created')
        permissions = (
            ('start_other_executor', _('Can start other executor')),
            ('stop_other_executor', _('Can stop other executor')),
        )

    def __str__(self):
        return '%s: %s' % (str(self.role), str(self.user))

    @property
    def time_total(self):
        """Returns the seconds spended for total work."""
        start = self.start
        stop = self.stop
        if start is None:
            return 0
        if stop is None:
            stop = now()
        return (stop - start).total_seconds()

    @property
    def time_work(self):
        """Returns the seconds spended for work on sprints."""
        time = self.time_sprints
        sprint = self.sprint
        if sprint:
            time += (now() - sprint).total_seconds()
        return time

    @property
    def is_current(self):
        return bool(self.sprint)

    def start_sprint(self, save=False):
        """Sets the start of the new sprint."""
        assert not self.sprint
        if not self.start:
            self.start = now()
        self.sprint = now()
        if save:
            self.save()
        return self.sprint

    def stop_sprint(self, save=False):
        """Stops the current sprint."""
        sprint = self.sprint
        assert sprint
        _now = now()
        time = (_now - sprint).total_seconds()
        self.time_sprints += time
        self.sprint = None
        self.stop = _now
        if save:
            self.save()
        return time



# Generated by Django 2.2 on 2019-04-18 10:01

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import proplan.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0011_update_proxy_permissions'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created')),
                ('updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='updated')),
                ('file', models.FileField(upload_to=proplan.models.upload_attach, verbose_name='file')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='proplan_attaches', to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'verbose_name': 'attach',
                'ordering': ('-updated', '-created'),
                'verbose_name_plural': 'attaches',
            },
        ),
        migrations.CreateModel(
            name='Comment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='updated')),
                ('message', models.TextField(db_index=True, verbose_name='message body')),
                ('changes', models.TextField(blank=True, editable=False, verbose_name='thread changes')),
                ('attachments', models.ManyToManyField(blank=True, to='proplan.Attachment', verbose_name='attachments')),
            ],
            options={
                'verbose_name': 'comment',
                'ordering': ('-updated', '-created'),
                'verbose_name_plural': 'comments',
            },
        ),
        migrations.CreateModel(
            name='Stage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=10, verbose_name='name')),
                ('description', models.TextField(blank=True, verbose_name='description')),
                ('is_finished', models.BooleanField(default=False, verbose_name='is finished')),
                ('is_abs', models.BooleanField(default=False, help_text='Only one record can be used for new stage for Automatic Bug System.', verbose_name='for automatic bug system')),
            ],
            options={
                'verbose_name': 'stage',
                'ordering': ('id',),
                'verbose_name_plural': 'stages',
            },
        ),
        migrations.CreateModel(
            name='Tracker',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=10, verbose_name='name')),
                ('description', models.TextField(blank=True, verbose_name='description')),
                ('is_abs', models.BooleanField(default=False, help_text='Only one record can be used for Automatic Bug System.', verbose_name='for automatic bug system')),
            ],
            options={
                'verbose_name': 'tracker',
                'ordering': ('id',),
                'verbose_name_plural': 'trackers',
            },
        ),
        migrations.CreateModel(
            name='Thread',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='updated')),
                ('priority', models.SmallIntegerField(choices=[(1, 'low'), (2, 'normal'), (3, 'high'), (4, 'urgent'), (5, 'immediate')], db_index=True, default=1, verbose_name='priority')),
                ('is_abs', models.BooleanField(db_index=True, default=False, verbose_name='from automatic bug system')),
                ('title', models.CharField(db_index=True, max_length=255, verbose_name='title')),
                ('message', models.TextField(blank=True, db_index=True, verbose_name='message body')),
                ('attachments', models.ManyToManyField(blank=True, to='proplan.Attachment', verbose_name='attachments')),
                ('author', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='proplan_author_threads', to=settings.AUTH_USER_MODEL, verbose_name='author')),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='children', to='proplan.Thread', verbose_name='parent thread')),
                ('stage', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='proplan.Stage', verbose_name='stage')),
                ('tracker', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='proplan.Tracker', verbose_name='tracker')),
            ],
            options={
                'verbose_name': 'thread',
                'ordering': ('-updated', '-created'),
                'verbose_name_plural': 'threads',
            },
        ),
        migrations.CreateModel(
            name='Role',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=50, verbose_name='name')),
                ('description', models.TextField(blank=True, verbose_name='description')),
                ('permissions', models.ManyToManyField(blank=True, limit_choices_to={'content_type__app_label': 'proplan'}, related_name='proplan_roles', to='auth.Permission', verbose_name='permissions')),
                ('users', models.ManyToManyField(blank=True, to=settings.AUTH_USER_MODEL, verbose_name='users')),
            ],
            options={
                'verbose_name': 'user role',
                'ordering': ('name',),
                'verbose_name_plural': 'user roles',
            },
        ),
        migrations.CreateModel(
            name='Executor',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='updated')),
                ('start', models.DateTimeField(blank=True, db_index=True, editable=False, null=True, verbose_name='start working')),
                ('stop', models.DateTimeField(blank=True, db_index=True, editable=False, null=True, verbose_name='stop working')),
                ('sprint', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='start of the sprint')),
                ('time_sprints', models.IntegerField(db_index=True, default=0, editable=False, verbose_name='time of the sprints')),
                ('comments', models.ManyToManyField(blank=True, to='proplan.Comment', verbose_name='comments')),
                ('role', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='executors', to='proplan.Role', verbose_name='user role')),
                ('thread', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='executors', to='proplan.Thread', verbose_name='thread')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='proplan_executors', to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'ordering': ('-updated', '-created'),
                'verbose_name_plural': 'executors',
                'verbose_name': 'executor',
                'permissions': (('start_other_executor', 'Can start other executor'), ('stop_other_executor', 'Can stop other executor')),
            },
        ),
        migrations.AddField(
            model_name='comment',
            name='thread',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='proplan.Thread', verbose_name='thread'),
        ),
        migrations.AddField(
            model_name='comment',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='proplan_comments', to=settings.AUTH_USER_MODEL, verbose_name='user'),
        ),
    ]



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from django.urls import path, include
from proplan.views import pages, api, auto

app_name = 'proplan'

abs_patterns = [
    path('upload/', auto.UploadView.as_view(), name='upload'),
    path('create/', auto.CreateView.as_view(), name='create'),
    path('', auto.check, name='check'),
]

api_patterns = [
    path('attachments/', api.AttachmentsView.as_view(), name='attachments'),
    path('attachments/<int:id>/', api.AttachmentView.as_view(), name='attachment'),
    path('trackers/', api.TrackersView.as_view(), name='trackers'),
    path('trackers/<int:id>/', api.TrackerView.as_view(), name='tracker'),
    path('stages/', api.StagesView.as_view(), name='stages'),
    path('stages/<int:id>/', api.StageView.as_view(), name='stage'),
    path('roles/', api.RolesView.as_view(), name='roles'),
    path('roles/<int:id>/', api.RoleView.as_view(), name='role'),
    path('threads/', api.ThreadsView.as_view(), name='threads'),
    path('threads/<int:id>/', api.ThreadView.as_view(), name='thread'),
    path('comments/', api.CommentsView.as_view(), name='comments'),
    path('comments/<int:id>/', api.CommentView.as_view(), name='comment'),
    path('executors/', api.ExecutorsView.as_view(), name='executors'),
    path('executors/<int:id>/', api.ExecutorView.as_view(), name='executor'),
    path('start/<int:id>/', api.StartView.as_view(), name='start'),
    path('stop/<int:id>/', api.StopView.as_view(), name='stop'),
]


pages_patterns = [
    # path('settings/', pages.SettingsView.as_view(), name='settings'),
    path('attachments/', pages.AttachmentsView.as_view(), name='attachments'),
    path('threads/', pages.ThreadsView.as_view(), name='threads'),
    path('', pages.IndexView.as_view(), name='index'),
]

urlpatterns = [
    path('api/', include((api_patterns, 'api'))),
    path('abs/', include((abs_patterns, 'abs'))),
    path('', include(pages_patterns)),
]



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from json import loads

from django.http import QueryDict, JsonResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from proplan import forms
from proplan.decorators import access_required, abs_required


def parse_params(request):
    if request.content_type == 'application/json':
        return loads(request.body.decode('utf-8'))
    method = request.method
    if method == 'GET':
        return request.GET
    elif method == 'POST':
        return request.POST
    return QueryDict(request.body)


@method_decorator(access_required, name='dispatch')
class AccessMixin:
    """Mixing for user access."""
    def dispatch(self, *args, **kwargs):
        return super(AccessMixin, self).dispatch(*args, **kwargs)


@method_decorator(csrf_exempt, name='dispatch')
@method_decorator(abs_required, name='dispatch')
class ABSMixin:
    """Mixing for Automatic Bug System access."""
    def dispatch(self, *args, **kwargs):
        return super(ABSMixin, self).dispatch(*args, **kwargs)


class BaseUploadView(View):
    """View for uploading attachments."""

    def upload_attachment(self, request):
        form = forms.AttachmentForm(request.user, files=request.FILES)
        if form.is_valid():
            attachment = form.save()
            data = {
                'id': attachment.id,
                'name': attachment.name,
                'url': attachment.url,
                'thumb': attachment.thumb_url,
            }
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)

    def post(self, request):
        return self.upload_attachment(request)



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import ugettext_lazy as _

# All next settings must be within the dictionary PROPLAN, when you
# define them in the file settings.py
conf = getattr(settings, 'PROPLAN', {})

# Function allowing access to the application.
ACCESS_FUNCTION = conf.get('ACCESS_FUNCTION', 'proplan.access.view_thread')

# ABS - Automatic Bug System.
# While there is no key, the system does not work.
ABS_KEY = str(conf.get('ABS_KEY', ''))
if ABS_KEY and len(ABS_KEY) < 12:
    raise ImproperlyConfigured(
        'Proplan ABS_KEY is not secure: %s' % ABS_KEY)
# The cookie name for checking the ABS key.
ABS_COOKIE_NAME = conf.get('ABS_COOKIE_NAME', 'proplanabs')

# Path to uploading files.
ATTACH_UPLOAD_PATH = conf.get(
    'ATTACH_UPLOAD_PATH', 'proplan/attaches/%(date)s/%(code)s/%(filename)s')
# The size of the thumbnails for attached images
ATTACH_THUMB_SIZE = conf.get('ATTACH_THUMB_SIZE', (300, 300))
# List of recognized image extensions to be previewed.
ATTACH_THUMB_EXTENSIONS = conf.get('ATTACH_THUMB_EXTENSIONS', [
    '.png', '.jpg', '.jpeg', '.bmp',
])

PRIORITIES = conf.get('PRIORITIES', [
    (1, _('low')),
    (2, _('normal')),
    (3, _('high')),
    (4, _('urgent')),
    (5, _('immediate')),
])



"""
Router for application.
"""


class ProplanRouter:
    """
    A router to control all database operations on models in this
    application.
    """
    def db_for_read(self, model, **hints):
        """
        Attempts to read proplan models go to proplan database.
        """
        if model._meta.app_label == 'proplan':
            return 'proplan'
        return None

    def db_for_write(self, model, **hints):
        """
        Attempts to write proplan models go to proplan database.
        """
        if model._meta.app_label == 'proplan':
            return 'proplan'
        return None

    def allow_relation(self, obj1, obj2, **hints):
        """
        Allow relations if a model in the proplan app is involved.
        """
        if obj1._meta.app_label == 'proplan' or \
           obj2._meta.app_label == 'proplan':
            return True
        return None

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        """
        Make sure the proplan app only appears in the 'proplan'
        database.
        """
        if app_label == 'proplan':
            return db == 'proplan'
        return None



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
"""
JSON REST API views for Automatic Bug System.
"""
from django.http import JsonResponse
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.cache import cache_page
from django.views.generic import View

from proplan import forms
from proplan.conf import ABS_KEY
from proplan.models import Tracker, Stage
from proplan.views.base import parse_params, ABSMixin, BaseUploadView


@cache_page(60 * 5)
def check(request):
    """Checking the ready ABS."""
    ready = bool(
        ABS_KEY and
        Tracker.objects.get_abs() and
        Stage.objects.get_abs()
    )
    if ready:
        message = _('The ABS is ready to accept errors.')
    else:
        message = _('The ABS is not ready to accept errors.')
    data = {
        'ready': ready,
        'message': message,
    }
    return JsonResponse(data)


class UploadView(ABSMixin, BaseUploadView):
    """Uploads attachments from Automatic Bug System."""
    pass


class CreateView(ABSMixin, View):
    """Posting threads from Automatic Bug System."""

    def post(self, request):
        form = forms.ABSThreadCreateForm(parse_params(request))
        if form.is_valid():
            thread = form.save()
            data = {
                'id': thread.id,
                'title': thread.title,
            }
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
# from django.utils.translation import ugettext_lazy as _
from django.views.generic import TemplateView

from proplan import forms
from proplan.models import Attachment, Thread
from proplan.logic import Controller
from proplan.views.base import AccessMixin


class IndexView(AccessMixin, TemplateView):
    template_name = 'proplan/index.html'


class AttachmentsView(AccessMixin, TemplateView):
    """List of uploaded attachments from user."""
    template_name = 'proplan/attachments.html'
    ctrl = Controller(Attachment)

    @property
    def extra_context(self):
        user = self.request.user
        form = forms.AttachmentForm(user)
        page, orders, filters = self.ctrl.get(self.request)
        ctx = {
            'page': page,
            'orders': orders,
            'filters': filters,
            'form': form,
        }
        return ctx


class ThreadsView(AccessMixin, TemplateView):
    """List threads for users."""
    template_name = 'proplan/threads.html'
    ctrl = Controller(Thread)

    @property
    def extra_context(self):
        page, orders, filters = self.ctrl.get(self.request)
        ctx = {
            'page': page,
            'orders': orders,
            'filters': filters,
        }
        return ctx



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
"""
JSON REST API views for users.
"""
from django.core.exceptions import PermissionDenied
from django.http import JsonResponse
# from django.utils.translation import ugettext_lazy as _
from django.shortcuts import get_object_or_404
from django.views.generic import View

from proplan import forms
from proplan.models import (
    Attachment, Tracker, Stage, Role, Thread, Comment, Executor,
)
from proplan.logic import Controller
from proplan.serializers import ExecutorSerializer, ThreadSerializer
from proplan.views.base import parse_params, AccessMixin, BaseUploadView


attachment_ctrl = Controller(Attachment)


class AttachmentsView(AccessMixin, BaseUploadView):
    """Uploads and views attachments."""

    def get(self, request):
        user = request.user
        if not user.has_perm('proplan.view_attachment'):
            raise PermissionDenied
        page, orders, filters = attachment_ctrl.get_serialized(request)
        data = {'page': page, 'orders': orders, 'filters': filters}
        return JsonResponse(data)

    def post(self, request):
        user = request.user
        if not user.has_perm('proplan.add_attachment'):
            raise PermissionDenied
        return self.upload_attachment(request)


class AttachmentView(AccessMixin, View):
    """Managing one attachment."""

    def get(self, request, id):
        user = request.user
        if not user.has_perm('proplan.view_attachment'):
            raise PermissionDenied
        obj = get_object_or_404(Attachment, id=id)
        data = attachment_ctrl.serializer(obj)
        return JsonResponse(data)

    def delete(self, request, id):
        user = request.user
        if not user.has_perm('proplan.delete_attachment'):
            raise PermissionDenied
        obj = get_object_or_404(Attachment, id=id)
        obj.delete()
        return JsonResponse({'id': id})


tracker_ctrl = Controller(Tracker)


class TrackersView(AccessMixin, View):
    """Getting and posting trackers."""

    def get(self, request):
        user = request.user
        if not user.has_perm('proplan.view_tracker'):
            raise PermissionDenied
        page, orders, filters = tracker_ctrl.get_serialized(request)
        data = {'page': page, 'orders': orders, 'filters': filters}
        return JsonResponse(data)

    def post(self, request):
        user = request.user
        if not user.has_perm('proplan.add_tracker'):
            raise PermissionDenied
        form = forms.TrackerForm(parse_params(request))
        if form.is_valid():
            tracker = form.save()
            data = tracker_ctrl.serializer(tracker)
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)


class TrackerView(AccessMixin, View):
    """Managing one tracker."""

    def get(self, request, id):
        user = request.user
        if not user.has_perm('proplan.view_tracker'):
            raise PermissionDenied
        obj = get_object_or_404(Tracker, id=id)
        data = tracker_ctrl.serializer(obj)
        return JsonResponse(data)

    def patch(self, request, id):
        user = request.user
        if not user.has_perm('proplan.change_tracker'):
            raise PermissionDenied
        obj = get_object_or_404(Tracker, id=id)
        form = forms.TrackerForm(parse_params(request), instance=obj)
        if form.is_valid():
            tracker = form.save()
            data = tracker_ctrl.serializer(tracker)
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)

    put = patch

    def delete(self, request, id):
        user = request.user
        if not user.has_perm('proplan.delete_tracker'):
            raise PermissionDenied
        obj = get_object_or_404(Tracker, id=id)
        obj.delete()
        return JsonResponse({'id': id})


stage_ctrl = Controller(Stage)


class StagesView(AccessMixin, View):
    """Getting and posting stages."""

    def get(self, request):
        user = request.user
        if not user.has_perm('proplan.view_stage'):
            raise PermissionDenied
        page, orders, filters = stage_ctrl.get_serialized(request)
        data = {'page': page, 'orders': orders, 'filters': filters}
        return JsonResponse(data)

    def post(self, request):
        user = request.user
        if not user.has_perm('proplan.add_stage'):
            raise PermissionDenied
        form = forms.StageForm(parse_params(request))
        if form.is_valid():
            stage = form.save()
            data = stage_ctrl.serializer(stage)
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)


class StageView(AccessMixin, View):
    """Managing one stage."""

    def get(self, request, id):
        user = request.user
        if not user.has_perm('proplan.view_stage'):
            raise PermissionDenied
        obj = get_object_or_404(Stage, id=id)
        data = stage_ctrl.serializer(obj)
        return JsonResponse(data)

    def patch(self, request, id):
        user = request.user
        if not user.has_perm('proplan.change_stage'):
            raise PermissionDenied
        obj = get_object_or_404(Stage, id=id)
        form = forms.StageForm(parse_params(request), instance=obj)
        if form.is_valid():
            stage = form.save()
            data = stage_ctrl.serializer(stage)
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)

    put = patch

    def delete(self, request, id):
        user = request.user
        if not user.has_perm('proplan.delete_stage'):
            raise PermissionDenied
        obj = get_object_or_404(Stage, id=id)
        obj.delete()
        return JsonResponse({'id': id})


role_ctrl = Controller(Role)


class RolesView(AccessMixin, View):
    """Getting and posting user roles."""

    def get(self, request):
        user = request.user
        if not user.has_perm('proplan.view_role'):
            raise PermissionDenied
        page, orders, filters = role_ctrl.get_serialized(request)
        data = {'page': page, 'orders': orders, 'filters': filters}
        return JsonResponse(data)

    def post(self, request):
        user = request.user
        if not user.has_perm('proplan.add_role'):
            raise PermissionDenied
        form = forms.RoleForm(parse_params(request))
        if form.is_valid():
            role = form.save()
            data = role_ctrl.serializer(role)
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)


class RoleView(AccessMixin, View):
    """Managing one user role."""

    def get(self, request, id):
        user = request.user
        if not user.has_perm('proplan.view_role'):
            raise PermissionDenied
        obj = get_object_or_404(Role, id=id)
        data = role_ctrl.serializer(obj)
        return JsonResponse(data)

    def patch(self, request, id):
        user = request.user
        if not user.has_perm('proplan.change_role'):
            raise PermissionDenied
        obj = get_object_or_404(Role, id=id)
        form = forms.RoleForm(parse_params(request), instance=obj)
        if form.is_valid():
            role = form.save()
            data = role_ctrl.serializer(role)
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)

    put = patch

    def delete(self, request, id):
        user = request.user
        if not user.has_perm('proplan.delete_role'):
            raise PermissionDenied
        obj = get_object_or_404(Role, id=id)
        obj.delete()
        return JsonResponse({'id': id})


thread_ctrl = Controller(Thread, serializer=ThreadSerializer)


class ThreadsView(AccessMixin, View):
    """Getting and posting threads."""

    def get(self, request):
        user = request.user
        if not user.has_perm('proplan.view_thread'):
            raise PermissionDenied
        page, orders, filters = thread_ctrl.get_serialized(request)
        data = {'page': page, 'orders': orders, 'filters': filters}
        return JsonResponse(data)

    def post(self, request):
        user = request.user
        if not user.has_perm('proplan.add_thread'):
            raise PermissionDenied
        form = forms.ThreadCreateForm(user, parse_params(request))
        if form.is_valid():
            thread = form.save()
            data = thread_ctrl.serializer(thread)
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)


class ThreadView(AccessMixin, View):
    """Managing one thread."""

    def get(self, request, id):
        user = request.user
        if not user.has_perm('proplan.view_thread'):
            raise PermissionDenied
        obj = get_object_or_404(Thread, id=id)
        data = thread_ctrl.serializer(obj)
        return JsonResponse(data)

    def patch(self, request, id):
        user = request.user
        if not user.has_perm('proplan.change_thread'):
            raise PermissionDenied
        obj = get_object_or_404(Thread, id=id)
        form = forms.ThreadChangeForm(
            user, parse_params(request), instance=obj)
        if form.is_valid():
            thread = form.save()
            # print(form.diff)
            data = thread_ctrl.serializer(thread)
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)

    put = patch

    def delete(self, request, id):
        user = request.user
        if not user.has_perm('proplan.delete_thread'):
            raise PermissionDenied
        obj = get_object_or_404(Thread, id=id)
        obj.delete()
        return JsonResponse({'id': id})


comment_ctrl = Controller(Comment)


class CommentsView(AccessMixin, View):
    """Getting and posting comments."""

    def get(self, request):
        user = request.user
        if not user.has_perm('proplan.view_comment'):
            raise PermissionDenied
        page, orders, filters = comment_ctrl.get_serialized(request)
        data = {'page': page, 'orders': orders, 'filters': filters}
        return JsonResponse(data)

    def post(self, request):
        user = request.user
        if not user.has_perm('proplan.add_comment'):
            raise PermissionDenied
        form = forms.CommentCreateForm(user, parse_params(request))
        if form.is_valid():
            comment = form.save()
            data = comment_ctrl.serializer(comment)
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)


class CommentView(AccessMixin, View):
    """Managing one comment."""

    def get(self, request, id):
        user = request.user
        if not user.has_perm('proplan.view_comment'):
            raise PermissionDenied
        obj = get_object_or_404(Comment, id=id)
        data = comment_ctrl.serializer(obj)
        return JsonResponse(data)

    def patch(self, request, id):
        user = request.user
        obj = get_object_or_404(Comment, id=id)
        if obj.user == user:
            pass
        elif not user.has_perm('proplan.change_comment'):
            raise PermissionDenied
        form = forms.CommentChangeForm(
            user, parse_params(request), instance=obj)
        if form.is_valid():
            comment = form.save()
            data = comment_ctrl.serializer(comment)
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)

    put = patch

    def delete(self, request, id):
        user = request.user
        if not user.has_perm('proplan.delete_comment'):
            raise PermissionDenied
        obj = get_object_or_404(Comment, id=id)
        obj.delete()
        return JsonResponse({'id': id})


executor_ctrl = Controller(Executor, serializer=ExecutorSerializer)


class ExecutorsView(AccessMixin, View):
    """Getting and posting executors."""

    def get(self, request):
        user = request.user
        if not user.has_perm('proplan.view_executor'):
            raise PermissionDenied
        page, orders, filters = executor_ctrl.get_serialized(request)
        data = {'page': page, 'orders': orders, 'filters': filters}
        return JsonResponse(data)

    def post(self, request):
        user = request.user
        if not user.has_perm('proplan.add_executor'):
            raise PermissionDenied
        form = forms.ExecutorCreateForm(user, parse_params(request))
        if form.is_valid():
            executor = form.save()
            data = executor_ctrl.serializer(executor)
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)


class ExecutorView(AccessMixin, View):
    """Managing one executor."""

    def get(self, request, id):
        user = request.user
        if not user.has_perm('proplan.view_executor'):
            raise PermissionDenied
        obj = get_object_or_404(Executor, id=id)
        data = executor_ctrl.serializer(obj)
        return JsonResponse(data)

    def patch(self, request, id):
        user = request.user
        obj = get_object_or_404(Executor, id=id)
        if obj.user == user:
            pass
        elif not user.has_perm('proplan.change_executor'):
            raise PermissionDenied
        form = forms.ExecutorChangeForm(
            user, parse_params(request), instance=obj)
        if form.is_valid():
            executor = form.save()
            data = executor_ctrl.serializer(executor)
            return JsonResponse(data)
        data = {'errors': form.errors.get_json_data(escape_html=True)}
        return JsonResponse(data, status=400)

    put = patch

    def delete(self, request, id):
        user = request.user
        if not user.has_perm('proplan.delete_executor'):
            raise PermissionDenied
        obj = get_object_or_404(Executor, id=id)
        obj.delete()
        return JsonResponse({'id': id})


class StartView(AccessMixin, View):
    """Starting executor work."""

    def post(self, request, id):
        user = request.user
        qs = Executor.objects.select_related('thread__stage')
        executor = get_object_or_404(qs, id=id)
        if executor.thread.stage.is_finished:
            raise PermissionDenied
        elif executor.user == user:
            pass
        elif not user.has_perm('proplan.start_other_executor'):
            raise PermissionDenied
        if not executor.sprint:
            executor.start_sprint(save=True)
        return JsonResponse({
            'id': executor.id,
            'start': executor.sprint,
            'time_total': executor.time_total,
            'time_work': executor.time_work,
        })


class StopView(AccessMixin, View):
    """Stop executor work."""

    def post(self, request, id):
        user = request.user
        qs = Executor.objects.select_related('thread__stage')
        executor = get_object_or_404(qs, id=id)
        if executor.user == user:
            pass
        elif not user.has_perm('proplan.stop_other_executor'):
            raise PermissionDenied
        if executor.sprint:
            executor.stop_sprint(save=True)
        return JsonResponse({
            'id': executor.id,
            'stop': executor.stop,
            'time_total': executor.time_total,
            'time_work': executor.time_work,
        })



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class DefaultConfig(AppConfig):
    name = 'proplan'
    verbose_name = _('ProPlan')



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from django.contrib.auth.models import Permission
from django.contrib.auth.backends import ModelBackend


class ExtraModelBackend(ModelBackend):

    def _get_role_permissions(self, user_obj):
        return Permission.objects.filter(proplan_roles__users=user_obj)

    def get_role_permissions(self, user_obj, obj=None):
        """
        Return a set of permission strings the user `user_obj` has from the
        roles they belong.
        """
        return self._get_permissions(user_obj, obj, 'role')

    def get_all_permissions(self, user_obj, obj=None):
        if not user_obj.is_active or user_obj.is_anonymous or obj is not None:
            return set()
        if not hasattr(user_obj, '_perm_cache'):
            user_obj._perm_cache = {
                *self.get_user_permissions(user_obj),
                *self.get_group_permissions(user_obj),
                *self.get_role_permissions(user_obj),
            }
        return user_obj._perm_cache



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from django.db.models import Model, Manager
from django.db.models.fields.files import FieldFile
from django.utils.encoding import force_text


def _instance_serializer(instance):
    return [
        instance.pk,
        force_text(instance),
    ]


class ObjectSerializer:
    """
    Serialize the fields of the object model.

    If the list of fields "only" is passed - it serializes only them if it is
    not present, then serializes all fields, except for the list "exclude".

    """
    model = None
    only = ()
    exclude = ()

    def __init__(self, model=None, only=None, exclude=None):
        if model is not None:
            self.model = model
        if only is not None:
            self.only = only
        elif exclude is not None:
            self.exclude = exclude
        self.is_ready = False
        if self.model:
            self.set_fields(self.model._meta)

    def __call__(self, instance, *args, **kwargs):
        return self.serialize(instance, *args, **kwargs)

    def set_fields(self, meta):
        # Collect all fields in the one heap.
        all = [f for f in meta.fields]
        if meta.many_to_many:
            all.extend([f for f in meta.many_to_many])
        # Make the list of needed fields.
        if self.only:
            list_fields = [f for f in all if f.name in self.only]
        elif self.exclude:
            list_fields = [f for f in all if f.name not in self.exclude]
        else:
            list_fields = all
        self.list_fields = list_fields
        self.is_ready = True
        return list_fields

    def serialize(self, instance, *args, **kwargs):
        """Full serialization the instance."""
        if not self.is_ready:
            self.set_fields(instance._meta)

        serializer = _instance_serializer
        result = {}
        fields = result

        for f in self.list_fields:
            name = f.name
            value = None
            data = getattr(instance, name, None)
            if isinstance(data, FieldFile):
                if data.name:
                    value = data.url
            elif isinstance(data, Model):
                value = serializer(data)
            elif isinstance(data, Manager):
                value = [serializer(i) for i in data.all()]
            else:
                value = data
            fields[name] = value
        return self.extend(instance=instance, data=result, *args, **kwargs)

    def extend(self, instance, data, *args, **kwargs):
        """The extension of the standard serialization."""
        return data


class ExecutorSerializer(ObjectSerializer):
    def extend(self, instance, data, *args, **kwargs):
        """The extension of the standard serialization."""
        data['time_total'] = instance.time_total
        data['time_work'] = instance.time_work
        data['is_current'] = instance.is_current
        return data


class ThreadSerializer(ObjectSerializer):
    def extend(self, instance, data, *args, **kwargs):
        """The extension of the standard serialization."""
        data['time_total'] = instance.time_total
        data['time_work'] = instance.time_work
        return data



# -*- coding: utf-8 -*-
import json

from django import forms
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from proplan import models
from proplan.logic import create_attachment, similarity_checking


class AttachmentForm(forms.Form):
    file = forms.FileField(required=True)

    def __init__(self, user, *args, **kwargs):
        self.user = user
        super(AttachmentForm, self).__init__(*args, **kwargs)

    def clean_file(self):
        file = self.cleaned_data['file']
        if not hasattr(file, 'content_type'):
            raise forms.ValidationError(_('Parameter "file" is not valid.'))
        name = file.name
        if '.' not in name:
            raise forms.ValidationError(_('File not contain extension.'))
        if '/' in name:
            raise forms.ValidationError(_('File name is contains slash.'))
        return file

    def save(self):
        file = self.cleaned_data['file']
        return create_attachment(self.user, file)


class TrackerForm(forms.ModelForm):

    class Meta:
        model = models.Tracker
        exclude = []


class StageForm(forms.ModelForm):

    class Meta:
        model = models.Stage
        exclude = []


class RoleForm(forms.ModelForm):

    class Meta:
        model = models.Role
        exclude = []


class CheckThreadMixing:

    def similarity_checking(self, cleaned_data):
        thread = self.instance
        other = similarity_checking(
            cleaned_data.get('title', thread.title),
            cleaned_data.get('message', thread.message),
            thread.is_abs,
            exclude_id=thread.id,
        )
        if other:
            raise forms.ValidationError(
                _('Similarity theads is exists: %(other)s.'),
                params={'other': other}
            )


class ThreadCreateForm(CheckThreadMixing, forms.ModelForm):

    class Meta:
        model = models.Thread
        fields = [
            'parent', 'tracker', 'stage', 'priority', 'title',
            'message', 'attachments',
        ]

    def __init__(self, user, *args, **kwargs):
        self.user = user
        if 'instance' not in kwargs:
            kwargs['instance'] = models.Thread(author=user, is_abs=False)
        super(ThreadCreateForm, self).__init__(*args, **kwargs)

    def clean(self):
        cleaned_data = super().clean()
        self.similarity_checking(cleaned_data)
        return cleaned_data


class ThreadChangeForm(CheckThreadMixing, forms.ModelForm):

    comment = forms.CharField(required=False)

    class Meta:
        model = models.Thread
        fields = [
            'parent', 'tracker', 'stage', 'priority', 'title',
            'message', 'attachments',
        ]

    def __init__(self, user, *args, **kwargs):
        self.user = user
        assert 'instance' in kwargs
        super(ThreadChangeForm, self).__init__(*args, **kwargs)

    def make_diff(self, data):
        self.diff = diff = {}
        instance = self.instance
        for name in self.changed_data:
            if name == 'comment':
                continue
            if name == 'attachments':
                old = list(instance.attachments.values_list('id', flat=True))
                new = data.get(name)
                if new is None:
                    new = []
                else:
                    new = list(new.values_list('id', flat=True))
                old.sort()
                new.sort()
            else:
                old = force_text(getattr(instance, name))
                new = force_text(data.get(name))
            diff[name] = [old, new]
        return diff

    def clean(self):
        cleaned_data = super().clean()
        self.similarity_checking(cleaned_data)
        self.make_diff(cleaned_data)
        return cleaned_data

    def save(self, *args, **kwargs):
        thread = super().save(*args, **kwargs)
        message = self.cleaned_data['comment']
        changes = json.dumps(self.diff)
        comment = models.Comment(
            user=self.user, thread=thread,
            message=message, changes=changes)
        comment.save()
        return thread


class ABSThreadCreateForm(CheckThreadMixing, forms.ModelForm):
    """
    Form for creating thread from Automaic Bug System.

    """

    class Meta:
        model = models.Thread
        fields = ['parent', 'title', 'message', 'attachments']

    def __init__(self, *args, **kwargs):
        if 'instance' not in kwargs:
            kwargs['instance'] = models.Thread(
                is_abs=True,
                tracker=models.Tracker.objects.get_abs(),
                stage=models.Stage.objects.get_abs(),
            )
        super(ABSThreadCreateForm, self).__init__(*args, **kwargs)

    def clean_tracker(self):
        tracker = self.cleaned_data['tracker']
        if not tracker.is_abs:
            raise forms.ValidationError(_('Tracker is not for ABS.'))
        return tracker

    def clean_stage(self):
        stage = self.cleaned_data['stage']
        if not stage.is_abs:
            raise forms.ValidationError(_('Stage is not for ABS.'))
        return stage

    def clean(self):
        instance = self.instance
        if not (instance.tracker and instance.stage):
            raise forms.ValidationError(
                _('The ABS is not ready to accept errors.'))
        return super().clean()


class CommentCreateForm(forms.ModelForm):

    class Meta:
        model = models.Comment
        fields = ['thread', 'message', 'attachments']

    def __init__(self, user, *args, **kwargs):
        self.user = user
        if 'instance' not in kwargs:
            kwargs['instance'] = models.Comment(user=user)
        super(CommentCreateForm, self).__init__(*args, **kwargs)


class CommentChangeForm(forms.ModelForm):

    class Meta:
        model = models.Comment
        fields = ['message', 'attachments']

    def __init__(self, user, *args, **kwargs):
        self.user = user
        assert 'instance' in kwargs
        super(CommentChangeForm, self).__init__(*args, **kwargs)


class ExecutorCreateForm(forms.ModelForm):

    class Meta:
        model = models.Executor
        fields = ['thread', 'user', 'role']

    def __init__(self, user, *args, **kwargs):
        self.user = user
        super(ExecutorCreateForm, self).__init__(*args, **kwargs)


class ExecutorChangeForm(forms.ModelForm):

    class Meta:
        model = models.Executor
        fields = ['role']

    def __init__(self, user, *args, **kwargs):
        self.user = user
        assert 'instance' in kwargs
        super(ExecutorChangeForm, self).__init__(*args, **kwargs)



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
"""
Functions for access control.
"""
from proplan.conf import ABS_COOKIE_NAME, ABS_KEY


def authenticated(user):
    """For authenticated users."""
    return user.is_active and user.is_authenticated


def staff(user):
    """For employers and superusers."""
    return authenticated(user) and (user.is_staff or user.is_superuser)


def superuser(user):
    """For superusers only."""
    return authenticated(user) and user.is_superuser


def view_thread(user):
    """For users with view permission for Thread model."""
    return authenticated(user) and user.has_perm('proplan.view_thread')


def check_abs_key(request):
    """Checks the request for Automatic Bug System."""
    key = request.COOKIES.get(ABS_COOKIE_NAME, '')
    if key == ABS_KEY:
        request.META['ABS_KEY'] = key
        return True
    if 'ABS_KEY' in request.META:
        request.META.pop('ABS_KEY')
    return False



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from setuptools import setup

# Dynamically calculate the version based on proplan.VERSION.
version = __import__('proplan').get_version()

with open('README.rst', 'r') as f:
    long_description = f.read()

setup(
    name='django-proplan',
    version=version,
    description=(
        'Django application to assign the tasks and planning of your project.'
    ),
    long_description=long_description,
    author='Grigoriy Kramarenko',
    author_email='root@rosix.ru',
    url='https://gitlab.com/djbaldey/django-proplan/',
    license='BSD License',
    platforms='any',
    zip_safe=False,
    packages=['proplan'],
    include_package_data=True,
    install_requires=[],
    classifiers=[
        # List of Classifiers: https://pypi.org/classifiers/
        'Development Status :: 1 - Planning',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Natural Language :: Russian',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from functools import wraps

from django.contrib.auth.decorators import user_passes_test
from django.core.exceptions import PermissionDenied
from django.utils.module_loading import import_string


from proplan.conf import ACCESS_FUNCTION
from proplan.access import check_abs_key

access = import_string(ACCESS_FUNCTION)


def access_required(function=None, login_url=None, raise_exception=False):
    """
    Decorator for views that checks whether a user access.
    """
    def check_perms(user):
        # First, check if the user has permission.
        if access(user):
            return True
        # If you need a 403 handler, throw an exception.
        if raise_exception:
            raise PermissionDenied
        # Or show the entry form.
        return False
    actual_decorator = user_passes_test(check_perms, login_url=login_url)
    if function:
        return actual_decorator(function)
    return actual_decorator


def abs_required(function=None):
    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):
            if request.user.is_authenticated or not check_abs_key(request):
                raise PermissionDenied
            return view_func(request, *args, **kwargs)
        return _wrapped_view
    if function:
        return decorator(function)
    return decorator



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from functools import reduce
from operator import or_ as OR
from PIL import Image, ImageOps

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.paginator import Paginator
from django.db.models import Q, CharField
from django.shortcuts import _get_queryset
# from django.utils.html import escape
from django.utils.translation import ugettext_lazy as _

from proplan.models import Attachment, Thread
from proplan.conf import ATTACH_THUMB_SIZE
from proplan.serializers import ObjectSerializer


User = get_user_model()


def create_attachment(user, file):
    """Creats the attachment file."""
    if user:
        assert isinstance(user, (User, AnonymousUser)), (
            'Parameter "user" is not valid.')
        if user.is_anonymous:
            user = None
    assert file and hasattr(file, 'content_type'), (
        'Parameter "file" is not valid.')
    name = file.name
    assert '.' in name, 'File not contain extension'
    assert '/' not in name, 'File contain slash'
    attachment = Attachment(user=user, file=file)
    attachment.full_clean()
    attachment.save()
    if attachment.use_thumbnail:
        try:
            with Image.open(attachment.path) as image:
                thumb = ImageOps.fit(image, ATTACH_THUMB_SIZE, Image.ANTIALIAS)
                # Save the thumbnail.
                thumb.save(attachment.thumb_path, 'png')
        except OSError as e:
            attachment.delete()
            raise TypeError(
                _('The image attachment is incorrect format.')
            )
    return attachment


def similarity_checking(title, message, is_abs=False, tracker=None,
                        exclude_id=None):
    qs = Thread.objects.all()
    qs = qs.none()
    if exclude_id:
        qs = qs.exclude(id=exclude_id)
    return list(qs.values_list('id', flat=True))


class Controller:
    """
    Filtering, ordering and pagination controller for QuerySet.
    """
    search_fields = None
    ordering_fields = None
    filtering_fields = None
    exclude_fields = ('password',)
    serializer = None

    def __init__(self, model, serializer=None):
        self.model = model
        meta = model._meta
        exclude = self.exclude_fields
        fields = [f for f in meta.fields if f.name not in exclude]
        if meta.many_to_many:
            fields.extend([
                f for f in meta.many_to_many if f.name not in exclude
            ])
        self.fields = fields
        if self.serializer is None:
            self.serializer = (
                serializer or ObjectSerializer(model=model, exclude=exclude)
            )
        if isinstance(self.serializer, type):
            self.serializer = self.serializer(model=model, exclude=exclude)

    def get_queryset(self):
        return _get_queryset(self.model)

    def get_search_fields(self):
        if self.search_fields is None:
            fields = [f.name for f in self.fields if isinstance(f, CharField)]
            # Search by models without text fields.
            if not fields:
                exclude = self.exclude_fields
                for field in self.fields:
                    rel = field.related_model
                    if rel:
                        prefix = field.name + '__%s'
                        for f in rel._meta.fields:
                            if isinstance(f, CharField):
                                fname = prefix % f.name
                                if fname not in exclude:
                                    fields.append(fname)
            self.search_fields = fields
        return self.search_fields

    def get_ordering_fields(self):
        if self.ordering_fields is None:
            self.ordering_fields = [f.name for f in self.fields if
                                    not f.related_model]
        return self.ordering_fields

    def get_filtering_fields(self):
        if self.filtering_fields is None:
            self.filtering_fields = [f.name for f in self.fields]
        return self.filtering_fields

    def ordering(self, queryset, ordering):
        """
        The function checks the collation and applies only the valid.
        """
        fields = self.get_ordering_fields()

        if not ordering or not fields:
            # Fix UnorderedObjectListWarning:
            if not getattr(queryset, 'ordered', True):
                queryset = queryset.order_by('pk')
            return queryset

        def valid(x):
            return bool(
                x and not x.startswith('--') and
                x.lstrip('-') in fields
            )

        if isinstance(ordering, str):
            if ordering.startswith('[') and ordering.endswith(']'):
                ordering = ordering[1:-1]
            ordering = ordering.split(',')
        ordering = [x for x in ordering if valid(x)]
        if ordering:
            queryset = queryset.order_by(*ordering)
        return queryset, ordering

    def search(self, queryset, query):
        """
        Returns a filtered set of data.

        Parameter `klass` may be Model, Manager or QuerySet object. If there
        is no list of fields 'fields' or there is no search string 'query',
        then returns the data set as is.
        """

        fields = self.get_search_fields()

        def construct_search(field_name):
            if field_name.startswith('^'):
                return "%s__istartswith" % field_name[1:]
            elif field_name.startswith('='):
                return "%s__iexact" % field_name[1:]
            elif field_name.startswith('@'):
                return "%s__search" % field_name[1:]
            else:
                return "%s__icontains" % field_name

        if fields:
            lookups = [construct_search(str(f)) for f in fields]
            if query not in ('', None, False, True):
                for bit in query.split():
                    queries = [Q(**{lookup: bit}) for lookup in lookups]
                    queryset = queryset.filter(reduce(OR, queries))

        return queryset

    def filtering(self, queryset, filters):
        """Filters the queryset."""

        if not filters:
            return queryset, {}

        fields = self.get_filtering_fields()

        def test_filtered(field):
            for f in fields:
                if field.startswith(f):
                    return True
            return False

        def test_inverse(s):
            return s.startswith('-')

        def test_bool(s, v):
            return s.endswith('__isnull') or v in ('true', 'false')

        def test_list(s):
            return s.endswith('__in') or s.endswith('__range')

        applied = {}

        for field, query in filters.items():
            if field == 'q':
                queryset = self.search(queryset, query)
                applied[field] = query
                continue

            if test_inverse(field):
                field = field[1:]
                func = queryset.exclude
            else:
                func = queryset.filter

            if not test_filtered(field):
                continue

            if isinstance(query, str):
                if query.startswith('[') and query.endswith(']'):
                    query = [x for x in query[1:-1].split(',') if x]
                elif test_list(field):
                    query = [x for x in query.split(',') if x]
                elif test_bool(field, query):
                    query = bool(query == 'true')
            queryset = func(Q(**{field: query}))
            applied[field] = query

        return queryset, applied

    def pagination(self, queryset, limit, page):
        """Returns paginator and page for queryset."""
        paginator = Paginator(queryset, limit)
        try:
            page = paginator.page(page)
        except paginator.EmptyPage:
            page = paginator.page(1)
        return paginator, page

    def get(self, request, queryset=None, page=1, limit=100, max_limit=1000):
        """
        Prepares the request and returns page, orders and filters.
        """
        data = request.GET.copy()
        if queryset is None:
            queryset = self.get_queryset()
        else:
            assert self.model == queryset.model
        if 'p' in data:
            try:
                page = int(data.pop('p'))
            except ValueError:
                pass
        if 'l' in data:
            try:
                limit = int(data.pop('l'))
            except ValueError:
                pass
        if limit > max_limit:
            limit = max_limit
        if 'o' in data:
            queryset, orders = self.ordering(queryset, data.pop('o'))
        else:
            orders = []
        queryset, filters = self.filtering(queryset, data)
        paginator, page = self.pagination(queryset, limit, page)
        return page, orders, filters

    def get_serialized(self, *args, **kwargs):
        serializer = kwargs.pop('serializer', None)
        page, orders, filters = self.get(*args, **kwargs)
        page = self.serialize_page(page, serializer)
        return page, orders, filters

    def serialize_page(self, page, serializer):
        if serializer is None:
            serializer = self.serializer
        paginator = page.paginator
        return {
            'objects': [serializer(obj) for obj in page.object_list],
            'number': page.number,
            'limit': paginator.per_page,
            'count': paginator.count,
            'pages': paginator.num_pages,
        }
